/* dc.c
 *
 * Sega Dreamcast specific functions
 */
#include <stdio.h>
#include <stdlib.h>
#include <kos.h>
#include "dc.h"
#include "io.h"
#include "gui.h"

/* Kallistios initialization flags */

KOS_INIT_FLAGS(INIT_DEFAULT);

pvr_poly_hdr_t polyHeader;
pvr_ptr_t frameBufferAddr;

int16 framebuffer[224*144];

/*  dc_init: Sets up a video mode, allocates memory
 *  for the framebuffer texture and initializes the PVR
 */

void dc_init() {
  pvr_poly_cxt_t cxt;

#ifdef DC_PAL
  vid_init(DM_640x480_PAL_IL, PM_RGB565);
#else
  vid_init(DM_640x480_NTSC_IL, PM_RGB565);
#endif

  pvr_init_params_t par = {
    { PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_0 },
    512*1024
  };

  pvr_init(&par);
  pvr_set_bg_color(0,0,0);

  frameBufferAddr = pvr_mem_malloc(256*256*2);
  memset(frameBufferAddr, 0, 256*256*2);


  pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY, 
		   (PVR_TXRFMT_ARGB1555 | PVR_TXRFMT_NONTWIDDLED),
		   256, 256 ,frameBufferAddr, PVR_FILTER_NONE);
  cxt.txr.alpha = 0;
  pvr_poly_compile(&polyHeader, &cxt);

  gui_setup();

}

/*  dc_shutdown: releases the framebuffer, shuts down the PVR
 */

void dc_shutdown() {
  pvr_mem_free(frameBufferAddr);
  gui_shutdown();
  pvr_shutdown();
}

uint8 dc_open_menu;

/*  dc_getControls: maps the DC controller buttons to the Wonderswan buttons.
 *  also checks whether user wants to access the menu.
 */

#define JOY_CENTER_X 128
#define JOY_CENTER_Y 128
#define JOY_THRESHOLD 20

void dc_setTimerHandler(void (*handler)(), int tps) {

  timer_prime(TMU2, tps, 1);
  irq_set_handler(EXC_TMU2_TUNI2, handler);
  timer_start(TMU2);
}

void dc_stopTimer() {
  timer_stop(TMU2);
}

void dc_getControls() {
  cont_cond_t contrState;
  uint8 dev;

  dev = maple_first_controller();
  cont_get_cond( dev, &contrState );

  ws_key_a_up = !(contrState.buttons & CONT_DPAD_UP) ? 1 :0;
  ws_key_a_down = !(contrState.buttons & CONT_DPAD_DOWN) ? 1:0;
  ws_key_a_left = !(contrState.buttons & CONT_DPAD_LEFT) ? 1:0;
  ws_key_a_right = !(contrState.buttons & CONT_DPAD_RIGHT) ? 1:0;
  ws_key_b_up = (contrState.joyy < (JOY_CENTER_Y - JOY_THRESHOLD)) ? 1:0;
  ws_key_b_down = (contrState.joyy > (JOY_CENTER_Y + JOY_THRESHOLD)) ? 1:0;
  ws_key_b_left = (contrState.joyx < (JOY_CENTER_X - JOY_THRESHOLD)) ? 1:0;
  ws_key_b_right = (contrState.joyx > (JOY_CENTER_X + JOY_THRESHOLD)) ? 1:0;
  ws_key_button_1 = !(contrState.buttons & CONT_A)? 1:0;
  ws_key_button_2 = !(contrState.buttons & CONT_B)? 1:0;
  ws_key_start = !(contrState.buttons & CONT_START)? 1:0;

  //Hack: Y button exits program
  if(!(contrState.buttons & CONT_Y))
    exit(0);

  dc_open_menu = !(contrState.buttons & CONT_X);
}

pvr_vertex_t displayVerts[4];

/*  dc_setDisplay: determines how the Wonderswan screen (i.e. the
 *  polygon with the framebuffer texture) will be displayed. The
 *  zoom factor determines the size, and if the boolean value
 *  rotated is true the screen is rotated 90 degrees (for games
 *  which use the Wonderswan's vertical mode)
 */

void dc_setDisplay( float zoom, int rotated ) {
#define TEX_U (224.0f/256.0f)
#define TEX_V (144.0f/256.0f)

  int i;
  int h = (int)(224.0f * zoom);
  int v = (int)(144.0f * zoom);

  for(i=0; i<4; i++) {
    displayVerts[i].flags = PVR_CMD_VERTEX;
    displayVerts[i].argb = 0xffffffff;
    displayVerts[i].oargb = 0x00000000;
    displayVerts[i].z = 1;
  }
  displayVerts[3].flags = PVR_CMD_VERTEX_EOL;

  displayVerts[0].u = 0.0f;  displayVerts[0].v = 0.0f;
  displayVerts[1].u = TEX_U; displayVerts[1].v = 0.0f;
  displayVerts[2].u = 0.0f;  displayVerts[2].v = TEX_V;
  displayVerts[3].u = TEX_U; displayVerts[3].v = TEX_V;

  if(rotated){
    displayVerts[0].x = displayVerts[1].x = (vid_mode->width - v) / 2;
    displayVerts[2].x = displayVerts[3].x = (vid_mode->width + v) / 2;
    displayVerts[1].y = displayVerts[3].y = (vid_mode->height - h) / 2;
    displayVerts[0].y = displayVerts[2].y = (vid_mode->height + h) / 2;
  } else {
    displayVerts[0].x = displayVerts[2].x = (vid_mode->width - h) / 2;
    displayVerts[1].x = displayVerts[3].x = (vid_mode->width + h) / 2;
    displayVerts[0].y = displayVerts[1].y = (vid_mode->height - v) / 2;
    displayVerts[2].y = displayVerts[3].y = (vid_mode->height + v) / 2;
  }
}

/*  dc_beginFrame: begin a frame */

void dc_beginFrame() {
  pvr_wait_ready();
  pvr_scene_begin();
}

/*  dc_endFrame: end a frame */

void dc_endFrame() {
  pvr_scene_finish();
}

/*  dc_dummyTransList: sends a dummmy translucent list to the PVR.
 *  This is necessary because there aren't any translucent polygons
 *  during normal operation, only when the menu is shown.
 */

void dc_dummyTransList() {
  pvr_list_begin(PVR_LIST_TR_POLY);
  gui_drawBox(0x00000000, 0x00000000, 1.0f, 1.0f, 2.0f, 2.0f, 1.0f);
  pvr_list_finish();
}

/*  fillTexture: copies the data from the framebuffer to the framebuffer
 *  texture.
 */

void fillTexture() {
  int i;
  int16 *from = framebuffer, *to = frameBufferAddr;

  for(i=0;i<144;i++) {
    memcpy(to, from, 224*2);
    from+=224; to+=256;
  }
}

/*  dc_blit: draws the framebuffer polygon. */

void dc_blit() {

  fillTexture();
  pvr_list_begin(PVR_LIST_OP_POLY);
  pvr_prim(&polyHeader, sizeof(polyHeader));

  pvr_prim(&displayVerts[0], sizeof(displayVerts[0]));
  pvr_prim(&displayVerts[1], sizeof(displayVerts[0]));
  pvr_prim(&displayVerts[2], sizeof(displayVerts[0]));
  pvr_prim(&displayVerts[3], sizeof(displayVerts[0]));

  pvr_list_finish();

}

/*  dc_getFrameBuffer: returns a pointer to the framebuffer
 */

int16 *dc_getFrameBuffer() {
  return framebuffer;//(int16 *) frameBufferAddr;
}

/*  timeGetTime: replaces the timeGetTime function in MSVC */

uint32 timeGetTime() {
  uint32 s, ms;
  timer_ms_gettime(&s, &ms);
  return s*1000+ms;
}

void write(FILE *fp, const void *buf, int size) {
  fwrite(buf, 1, size, fp);
}

void read(FILE *f, void* buf, int size) {
  fread(buf, 1, size, f);
}
